"""
The core specutils data objects package. This contains the
`~astropy.nddata.NDData`-inherited classes used for storing the spectrum data.
"""
from .spectrum import *  # noqa
from .spectral_region import *  # noqa
from .spectrum_collection import *  # noqa
from .spectrum_list import * # noqa
from .spectral_axis import *  # noqa
